module RayCalc (
    Q,Float3,
    getColor, normalizeV
   )
    where



--------------------------------------------------------------------------------
------------------------------- Math Definitions -------------------------------
-------------------------------------- Q ---------------------------------------

import Test.QuickCheck
import Debug.Trace

-- quaternion type
type Q = (Double, Double, Double, Double)

-- too much hassle to use Haskell's type declarations

-- multiply two quaternions
mult :: Q -> Q -> Q
mult (a1,b1,c1,d1) (a2,b2,c2,d2) = (x,i,j,k)
	where
		x = a1*a2 - b1*b2 - c1*c2 - d1*d2
		i = a1*b2 + a2*b1 + c1*d2 - c2*d1
		j = a1*c2 + a2*c1 + d1*b2 - d2*b1
		k = a1*d2 + a2*d1 + b1*c2 - b2*c1
		
--square a quaternion
--optimized mult
sq :: Q -> Q
sq (a,b,c,d) = (x,i,j,k)
	where
		x = a*a - b*b - c*c - d*d
		i = 2*a*b
		j = 2*a*c
		k = 2*a*d
		
-- multiply quaternion by scalar
-- optimized mult
smult :: Double -> Q -> Q
smult s (a,b,c,d) = (a*s,b*s,c*s,d*s)

-- add two quaternions
add :: Q -> Q -> Q
add (a,b,c,d) (w,x,y,z) = (a+w,b+x,c+y,d+z)

-- substract two quats
sub :: Q -> Q -> Q
sub (a,b,c,d) (w,x,y,z) = (a-w,b-x,c-y,d-z)

-- gets quaternion's length
len :: Q -> Double
len = sqrt . len2

-- get square of quaternion's length
len2 :: Q -> Double
len2 (a,b,c,d) = a*a + b*b + c*c + d*d

----------------------------------- Float3 -------------------------------------

type Float3 = (Double, Double, Double)	--3d vector

-- Float3 to Quaternion
f2q :: Float3 -> Q
f2q (a,b,c) = (a,b,c,0)

-- double to float3
-- used for (double * float3) operations
d2f :: Double -> Float3
d2f x = (x,x,x)

-- defines (+), (-), (*) for 3d vectors
-- where (*) is just pairwise multiplication
instance (Num a, Num b, Num c) => Num (a,b,c) where
	(x, y, z) + (x', y', z') = (x+x', y+y', z+z')
	(x, y, z) - (x', y', z') = (x-x', y-y', z-z')
	(x, y, z) * (x', y', z') = (x*x', y*y', z*z')
	fromInteger q = (fromInteger q, fromInteger q, fromInteger q)
	abs (x,y,z) = (abs x,abs y,abs z)
	signum (x,y,z) = (0,0,0)

--calculates dot product of two 3d vectors
dot :: Float3 -> Float3 -> Double
dot (a,b,c) (d,e,f) = a*d + b*e + c*f

-- normalizes a vector
normalizeV :: Float3 -> Float3
normalizeV (x,y,z) = (x/rt, y/rt, z/rt)
	where
		rt = sqrt(x*x + y*y + z*z)		-- square of dot of the vector

		
		
		
--------------------------------------------------------------------------------
-------------------------------- The Real Code! --------------------------------
--------------------------------------------------------------------------------


-- Calculates point normal 
-- used for coloring
--
-- PARAMETERS:
-- 		(x,y,z) - point for which the normal is calculated
-- 		c - julia set constant
-- 		n - maximum number of iterations
-- RETURNS:
-- 		normal of the point
delta = 0.001 -- distance from p to the points used for norm calc	
norm :: Float3 -> Q -> Int -> Float3
norm (x,y,z) c n = normalizeV (grX, grY, grZ)
	where
		pxm = (x-delta,y,z,0)
		pxp = (x+delta,y,z,0)
		pym = (x,y-delta,z,0)	-- points (roughly) defining the norm
		pyp = (x,y+delta,z,0)
		pzm = (x,y,z-delta,0)
		pzp = (x,y,z+delta,0)
		grX = getDist(calcZ n pxp c) - getDist(calcZ n pxm c)	--gradX,Y,Z
		grY = getDist(calcZ n pyp c) - getDist(calcZ n pym c)
		grZ = getDist(calcZ n pzp c) - getDist(calcZ n pzm c)

		
-- Calculates color for given point using its normal
-- Kinda modified Phong equation
-- PARAMETERS:
-- 		cam - camera location in 3d plane
-- 		l 	- location of light source
-- 		p   - the point we are wondering about
-- 		n   - the point's normal
-- RETURNS:
-- 		color of the point

diffColor = (1, 0.7, 0) 	-- base color
e = 10				 		-- shininess
specRefl = 0.45				-- specular reflection constant
-- maxdist = 
illum :: Float3 -> Float3 -> Float3 -> Float3 -> (Double -> Float3) -> Float3
illum cam l p n cFunc = ret
	where
		
		lv = normalizeV (l-p)
		cv = normalizeV (cam-p)
		nl = dot n lv
		rv = lv - d2f(2*nl) * n
		distcp = len ( f2q (cam - p))
		dc = cFunc ((distcp-3) / 6)
		ret = dc * d2f(max 0 nl) + d2f(specRefl * ((max (dot cv rv) 0) ^ e))
	


-- Returns intersection (new origin) of ray with given origin,distance
-- with the bounding sphere (that fully contains the set)
-- used to quickly move ray's origin vector near the julia set
-- PARAMETERS:
-- 		rayO - ray origin vector
-- 		rayD - ray direction vector
-- RETURNS:
-- 		new ray origin at the bounding sphere (if it intersects it)
--      does it actually intersect the sphere
sphereRadius = 3
intSphere :: Float3 -> Float3 -> (Float3, Bool)
intSphere rayO rayD = 
					if (disc >= 0) then 	--if the ray intersects the sphere
						let
							x0 = (disc-b)
							x1 = (-b-disc)
							x = (min x0 x1)
						in
							-- move the ray dist along the direction vector
							((rayO + d2f(x) * rayD), True)
					else
						(0, False)	--no intersection
	where
		b = dot rayO rayD
		c = (dot rayO rayO) - sphereRadius
		disc = sqrt(b*b-c)
		
		

-- gets point of intersection of a ray with the julia set 
-- uses the derivative of z (zp) and the distance estimator to move the ray
-- along its direction with conservative stepping
-- PARAMETERS:
-- 		maxIt      - max number of iterations to determine the set
-- 		eps        - distance to the set at which to accept we're "at" it
-- 		rayO, rayD - ray origin and direction vectors
-- 		c 		   - julia set constant we're iterating for
-- RETURN:
-- 			distance to the set (so we know if ray doesn't intersect it)
-- 			new ray origin
getIntersect :: Int -> Double -> Float3 -> Float3 -> Q -> (Double, Float3)
getIntersect maxIt eps rayO rayD c | dist <= eps || dot rayO rayO >= sphereRadius = 
										-- either we're close enough to the set
										-- or we are outside the bounding sphere (i.e. no intersection)
											(dist, rayO)
								   | otherwise = 
								   		-- move vector with dist along direction vector
								   		-- and re-do the approximation
								   		getIntersect maxIt eps (rayO + d2f(dist) * rayD) rayD c
	where
		(z,zp) = calcZ maxIt (f2q rayO) c
		dist = getDist (z,zp)
		

		
-- calculates estimate distance from a point to the nearest point on the set
-- using zn and z'n <the derivative of zn> (already iterated)
--
-- PARAMETERS:
-- 		(z, z')
-- RETURNS:
-- 		distance to the nearest point on the julia set 
getDist :: (Q, Q) -> Double
getDist (z, zp) = 0.5 * lz / (len zp) * (log lz)
	where
		lz = len z

		
		
-- calculates z, z' for given number of iterations, point, constant
-- iterates z(n+1) = z(n)^2 + c (TODO: allow different formulas)
-- and calculates its derivative, z', at the same time
-- breaks when we reach the treshold (i.e. z(n) diverges to infinity)
-- or continues till max iterations (n) meaning point is 'probably' in the set
-- bigger n means greater detail but much more computationaly expensive
--
-- PARAMETERS:
-- 		number of iterations
-- 		point for which we iterate
-- 		Julia set constant (defining the set)
-- RETURNS:
--		z(n) - the value of z after the iterations
--		z'(n) - the 'derivative' of z(n)
calcZ :: Int -> Q -> Q -> (Q, Q)
calcZ on z0 c = (x,y)
	where
		(x,y,z) = calcH on
		calcH 0 = (z0, (1,0,0,0), True)
		calcH n =
						let
					 		(oz, ozp, oc) = calcH (n-1)	-- old z, z'
						in	
					 		if oc then	-- if it still could be in the set 
					 					-- calculate for next n
							  let
								z = (sq oz) `add` c				-- new z
								zp = (2 `smult` oz) `mult` ozp	-- new z'
								cont = (len2 z < 4)				-- part of set or not
							  in
							  	(z, zp, cont)
							else 		-- just pass on the results
					 		  	(oz,ozp, False)
					 		  	
					 		 
					 		  	
-- Ray traces a ray until it hits the set and then calculates the color at
-- that position as seen from some eye and illuminated by some light. 
-- PARAMETERS:
-- 		rayO		- ray origin
--		rayD		- ray direction
--		cPos		- camera (eye) position
--		lightPos	- light source position
--		c			- constant for which the set is evaluated
--		eps			- detail of the image (epsilon)
--		n			- number of iterations
--		renShadows	- whether to use a (simple) shadowing algorithm for better coloring
--		colFunc		- the function that returns the base color of a pixel
-- RETURNS:
-- 		Color of the point along the ray
getColor :: Float3 -> Float3 -> Float3 -> Float3 -> Q -> Double -> Int -> Bool -> (Double -> Float3) -> Float3
getColor rO rD cP lP c eps maxIt rs cFunc = 
	let
		bc = (0.1,0.1,0.1)				-- back color
		--nrD = normalizeV rD				-- normalize direction
		(trO,intSp) = intSphere rO rD
		sprO = trO + rD * d2f(3*eps)	-- move a bit in the sphere 
										-- to account for floating point error
	in
		if (intSp) then
			let
				(dist,nrO) = (getIntersect maxIt eps sprO rD c)
			in
				if (dist <= eps) then
					let
						nrm = (norm nrO c maxIt)
						col = illum cP lP nrO nrm cFunc
					in
						-- simple shadow rendering algorithm
						if rs then
							let	
								-- ray from our point to light
								lr = normalizeV (lP - nrO)
								-- move it a bit towards light
								srO = nrO + d2f(3*eps)*lr
								-- does it intersect the set again
								(sd,_) = getIntersect maxIt eps srO lr c
							in
								-- if it does then it should be shadowed
								if(sd <= eps) then
									col * (d2f 0.5)	-- make it darker
								else
									col				-- normal color
						else
							col	--no shadow rendering: print normal color
				else	
					bc	-- there's no intersection with the set
		else	
			bc	-- there's no intersection with the sphere








