
import qualified Graphics.Rendering.OpenGL as GL
import Graphics.UI.GLUT
import Control.Monad
import RayCalc
import Data.List
import GHC.Float
import Debug.Trace
import Data.Array ((!), array, Array)
import System.Environment (getArgs)

type Point = (Int, Int)

--------------------------------------------------------------------------------
---------------------------------- CLI / Main ----------------------------------
--------------------------------------------------------------------------------

-- PARAMETERS:
-- 		rendered size
--		camera position
--		constant
--		iterations
--		specific color?

--		main.exe [-s render_size] [-cp camera_position] [-c julia_constant] 
-- 				 [-it num_iterations] [-col fl3]

-- parse CL with some default values
parseCL :: [String] -> (Int, Char, Q, Int, Float3, Bool) -> (Int, Char, Q, Int, Float3, Bool)
parseCL []  c = c
parseCL [x] (rs, pos, c, i, col, ns) | x == "-ns" = (rs, pos, c, i, col, False)
									 | otherwise = error "Invalid arguments"
parseCL (x:y:xs) (rs,pos,c,i,col, ns) | x ==  "-s"  = parseCL xs (read y, pos, c, i, col, ns)
			   					  | x == "-cp"  = parseCL xs (rs, head y, c, i, col, ns)
			   					  | x == "-c"   = parseCL xs (rs, pos, parseQ y, i, col, ns)
			   					  | x == "-it"  = parseCL xs (rs, pos, c, read y, col, ns)
			   					  | x == "-col" = parseCL xs (rs, pos, c, i, parseF3 y, ns)
			   					  | x == "-ns"  = parseCL (y:xs) (rs, pos, c, i, col, False)
			   					  | otherwise = error "Invalid arguments"
			   					 
-- don't remember if there's such a func alrdy declared
-- but no probs, writing it again as I couldn't find it
split :: Char -> String -> [String]
split _ [] = []
split c (s:xs) | s == c = split c xs
			   | otherwise = (s : takeWhile (/= c) xs) : split c (dropWhile (/= c) xs)
			   					
-- strings of type "(x,y,z)" where x,y,z are numbers
parseF3 :: String -> Float3
parseF3 (x:xy) = (a,b,c)
	where
		a = read (ints !! 0)
		b = read (ints !! 1)
		c = read (ints !! 2)
		ints = split ',' xys
		xys = take ((length xy) - 1) xy

-- same thing but for 4-tuples
parseQ :: String -> Q
parseQ (x:xy) = (a,b,c,d)
	where
		a = read (ints !! 0)
		b = read (ints !! 1)
		c = read (ints !! 2)
		d = read (ints !! 3)
		ints = split ',' xys
		xys = take ((length xy) - 1) xy

-- determines whether we use distance-based coloring
-- or color selected by user
isZero :: Float3 -> Bool
isZero (a,b,c) = (a == 0) && (b == 0) && (c == 0)

main :: IO()
main = do
	args <- getArgs
	let (size, cp, c, it, col, rs) = parseCL args (400, 'z', (-0.2, 0.8, 0, 0), 10, (0,0,0), True)
	display size cp c it col rs

--------------------------------------------------------------------------------
---------------------------------- Rendering -----------------------------------
--------------------------------------------------------------------------------

-- projected screen half-size
-- constant
psSize :: Double
psSize = 1.2

-- generates all points on the surface for given width/height
-- vals are [-1, 1]
myPoints :: Int -> Int -> [(GLfloat,GLfloat,GLfloat)]
myPoints w h = [(( 2 * (fromIntegral iw) / (fromIntegral w)) - 1, 2 * (fromIntegral ih/fromIntegral h) - 1, 0.0) | iw <- [1..w], ih <- [1..h]]


-- clumsy rainbow palette array
-- should be O(1) for element retrieval, right?
palette :: Double -> Float3
palette n = cpal ! (floor (n * 599))
cpal :: Array Int Float3
cpal = array (0,599) (c1 ++ c2 ++ c3 ++ c4 ++ c5 ++ c6)
	where
		c1 = [(truncate (i)		 , (1, i/100, 0)	  ) | i <- [0..99]]
		c2 = [(truncate (i + 100), ((100-i)/100, 1, 0)) | i <- [0..99]]
		c3 = [(truncate (i + 200), (0, 1, i/100)	  ) | i <- [0..99]]
		c4 = [(truncate (i + 300), (0, (100-i)/100, 1)) | i <- [0..99]]
		c5 = [(truncate (i + 400), (i/100, 0, 1)	  ) | i <- [0..99]]
		c6 = [(truncate (i + 500), (1, 0, (100-i)/100)) | i <- [0..99]]


-- calculates the color for a given point
--			  	x			y	   rOrig      camPos    lPos      c   epsilon   maxIt  renSh		color func
calcColor :: GLfloat -> GLfloat -> Float3  -> Float3 -> Float3 -> Q -> Double -> Int -> Bool -> (Double -> Float3) -> Char -> Color3 GLfloat
calcColor x y rOrig camPos lPos c eps maxIt renSh cFunc ccp = f2c(colF3)
	where
		rDirO = (getDir ccp) - rOrig
		rDir = (normalizeV rDirO)
		colF3 = (getColor rOrig rDir camPos lPos c eps maxIt renSh cFunc)
		getDir 'x' = (-3, psSize * float2Double x, psSize * float2Double y)
		getDir 'y' = (psSize * float2Double x, -3, psSize * float2Double y)
		getDir 'z' = (psSize * float2Double x, psSize * float2Double y, -3)
	

-- float3 to color3
f2c :: Float3 -> Color3 GLfloat
f2c (x,y,z) = Color3 ((double2Float x)::GLfloat) ((double2Float y) :: GLfloat) ((double2Float z) :: GLfloat)

-- simplified display
display :: Int -> Char -> Q -> Int -> Float3 -> Bool -> IO()
display sSize ccPos c maxIt col renSh = displayEx sSize cam cam l c 0.001 maxIt renSh (getPal col) ccPos
	where
		cam = (getCam ccPos) :: Float3
		l = (cam - (2,2,2)) :: Float3

getPal col | isZero col = palette
		   | otherwise = (\d -> col)
getCam c | c == 'x' = (-6, 0, 0)
		 | c == 'y' = (0, -6, 0)
		 | c == 'z' = (0, 0, -6)
		 | otherwise = error "Invalid camera dimension"

-- full display function (not implemented in CLI)
--			sSize	rOrig      camPos    lPos      c   epsilon   maxIt  renSh      color func
displayEx :: Int -> Float3  -> Float3 -> Float3 -> Q -> Double -> Int -> Bool -> (Double -> Float3) -> Char -> IO ()
displayEx sSize rOrig camPos lPos c eps maxIt renSh cFunc ccp = do
  	initialWindowSize  $= Size (fromIntegral sSize) (fromIntegral sSize)
  	(progname, _) <- getArgsAndInitialize
  	createWindow ("Julia Quaternion for c = " ++ show c)
  	displayCallback $= displaySub sSize rOrig camPos lPos c eps maxIt renSh cFunc ccp
  	mainLoop
  
-- display callback
displaySub sSize rOrig camPos lPos c eps maxIt renSh cFunc ccp = do
	clear [ColorBuffer]
	renderPrimitive Points $ 
  		sequence_  (concat [[color $ calcColor x y rOrig camPos lPos c eps maxIt renSh cFunc ccp, vertex$Vertex3 x y z]  | (x,y,z) <- (myPoints (fromIntegral sSize) (fromIntegral sSize))])
  	flush

  

  
  
  
  
  
  
  
